import logging
from collections import OrderedDict
from IPython import get_ipython

from .fly1d import Fly1D
from .fly2d import Fly2D

logger = logging.getLogger(__name__)

ip = get_ipython()

try:
    zpssx = ip.user_ns['zpssx']
    zpssy = ip.user_ns['zpssy']
    zpssz = ip.user_ns['zpssz']
except KeyError as ex:
    logger.error('Ensure that fly scan motor "%s" is still in the startup '
                 'configuration with the same name.', ex,
                 exc_info=ex)
    raise

coord_sys = 6
prog_num = 50
axes = OrderedDict([('x', dict(positioner=zpssx, axis_number=6)),
                    ('y', dict(positioner=zpssy, axis_number=7)),
                    ('z', dict(positioner=zpssz, axis_number=8))]
                    )
# change return_speed from 25.0 to 40.0; change dead_time from 0.002 to 0.005. (H. Yan, 01/10/2017)
configure_defaults = dict(return_speed=40.0,
                          dead_time=0.002,
                          fly_type='soft',
                          max_points=16384,
                          )


class Fly1D_ZP(Fly1D):
    def __init__(self, cmd='fly1d', **kwargs):
        super().__init__(axes=axes, coord_sys=coord_sys, prog_num=prog_num,
                         cmd=cmd, configure_defaults=configure_defaults,
                         **kwargs)


class Fly2D_ZP(Fly2D):
    def __init__(self, cmd='fly2d', **kwargs):
        super().__init__(axes=axes, coord_sys=coord_sys, prog_num=prog_num,
                         cmd=cmd, configure_defaults=configure_defaults,
                         **kwargs)



import asyncio
import logging
import IPython
from collections import OrderedDict

from .liveplot import FlyLivePlot

loop = asyncio.get_event_loop()
logger = logging.getLogger(__name__)


def get_default_xspress3():
    ip = IPython.get_ipython()
    return ip.user_ns['xspress3']


def signals_from_roi_name(roi_name, *, use_sum=False,
                          channels=None, xspress3=None,
                          channel_delimiter='_'):
    '''Get all ROI signals that match roi_name

    Parameters
    ----------
    roi_name : str
        Name of ROI
    use_sum : bool, optional
        Use the ROI value or the running sum
    channels : list, optional
        List of allowable channels to match
    xspress3 : Xspress3Detector, optional
        The detector instance. If unspecified, this is retrieved through the
        IPython user namespace variable 'xspress3'
    channel_delimiter : str, optional
        Rois are named in the form so as not to clash with other keys:
            Detector<-DELIMITER->ROI_NAME
        This function will match either Det1_ROI_NAME or ROI_NAME.
    '''

    if xspress3 is None:
        xspress3 = get_default_xspress3()

    signals = {}
    for roi in xspress3.enabled_rois:
        if channel_delimiter is not None and channel_delimiter in roi.name:
            name = roi.name.split(channel_delimiter, 1)[1]
        else:
            name = roi.name

        if name == roi_name or roi.name == roi_name:
            if channels is not None and roi.channel_num not in channels:
                continue

            if use_sum:
                # TODO
                # old version is ARRSUM1:ArrayData
                pass

            signals[roi.channel_num] = roi.settings.array_data
            # TODO: these data keys need to be the same...
            # TODO: if this is ever used in acquisition, it could be a problem
            signals[roi.channel_num].name = roi.name

    return [sig for chan, sig in
            sorted(signals.items(), key=lambda chan_sig: chan_sig[0])]


class FlyRoiPlot(FlyLivePlot):
    def __init__(self, roi_names, *, channels=None, use_sum=False,
                 group_rois=True, xspress3=None, **kwargs):

        if xspress3 is None:
            xspress3 = get_default_xspress3()

        if channels is None:
            channels = [1, 2, 3]

        self._channels = channels
        self._use_sum = use_sum
        self._group_rois = group_rois

        if isinstance(roi_names, str):
            roi_names = [roi_names]

        signals = OrderedDict()
        if group_rois:
            # group all channels together
            for roi_name in roi_names:
                signals[roi_name] = list(signals_from_roi_name(
                    roi_name, use_sum=use_sum, channels=channels))
        else:
            # keep all rois separate, use their full name
            for roi_name in roi_names:
                for sig in signals_from_roi_name(roi_name, use_sum=use_sum,
                                                 channels=channels):
                    signals[sig.name] = [sig]

        super().__init__(signals, point_signal=xspress3.settings.array_counter,
                         **kwargs)



import numpy as np

from .liveimage import FlyLiveImage
from .flydata import catch_exceptions

# from xray_vision.backend.mpl.cross_section_2d import CrossSection2DView
from xray_vision.qt_widgets import CrossSectionMainWindow
from matplotlib.backends.backend_qt5 import _create_qApp


def _new_window(title):
    fake_data = np.zeros((5, 5))
    return CrossSectionMainWindow(data_list=[fake_data], key_list=['fake'],
                                  title=title)


class FlyLiveCrossSection(FlyLiveImage):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        _create_qApp()
        if len(self.signals) != 1:
            raise ValueError('Cross section view only supports a single '
                             'ROI')

        self.label = list(self.signals.keys())[0]
        self.live_window = _new_window('Live update - {}'.format(self.label))
        self._final_window = None

    @property
    def final_window(self):
        if self._final_window is None:
            title = 'Scan {}: {}'.format(self.scan_id, self.label)
            self._final_window = _new_window(title)

        return self._final_window

    @catch_exceptions
    def plot(self, fig, data, *, final=False, **kwargs):
        window = self.live_window
        if final:
            window = self.final_window

        view = window._messenger._view._xsection
        view.update_image(data[self.label][::-1])
        window.show()

    def draw(self, fig=None):
        pass



import pytest

import numpy as np
import pandas as pd

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

import filestore
import filestore.api
import filestore.handlers

from hxnfly.fly1d import Fly1D
from hxnfly.bs import FlyPlan1D
from .fly_mock import (MockDetector, MockSignal)
from .sim_detector import TestDetector
from .fixtures import *


def make_1d_data(startx, endx, npts, gathered_points=100,
                 gate_enable='low_to_high'):
    time = np.linspace(0, 10.0, gathered_points)
    px = np.linspace(startx, endx, gathered_points)
    py = np.random.rand(gathered_points) / 10.0
    pz = np.random.rand(gathered_points) / 10.0

    enable = []
    exposed_count = (gathered_points // npts) - 1

    for i in range(npts):
        if gate_enable == 'low_to_high':
            enable.extend([0] + [1] * exposed_count)
        else:
            enable.extend([1] + [0] * exposed_count)

    if gate_enable == 'low_to_high':
        enable[-1] = 0
    else:
        enable[-1] = 1

    return [time, enable, px, py, pz]


@pytest.fixture(scope='function',
                params=['with_mock_detector', 'with_sim_detector',
                        'relative'])
def fly1d(request, monkeypatch, ppmac, axes, positioners, sim_det,
          ipython, global_state):
    if request.param == 'with_sim_detector':
        sim_det.count_time.put(0.01)
        detectors = [sim_det]
    else:
        detectors = [MockDetector('', name='det')]

    scan = Fly1D(axes=axes, positioners=positioners, detectors=detectors)
    scan.relative = (request.param == 'relative')

    def FakeClass(*args, **kwargs):
        print('initialized with', args, kwargs)
        scan.configure_defaults = dict(return_speed=5.0, dead_time=0.007,
                                       fly_type='soft', max_points=16384)
        return scan

    testx = positioners['testx']
    FlyPlan1D.scans = {frozenset({testx}): FakeClass,
                       }
    return scan


def setup_scan(fly1d, ipython, positioners, *, configure=True):
    startx, endx, npts = -1.0, 1.0, 2
    exposure_time = 1.0

    if configure:
        fly1d.configure(positioners['testx'], startx, endx, npts,
                        exposure_time=exposure_time, relative=fly1d.relative)

    sclr1 = ipython.user_ns['sclr1']
    sclr1.mca_by_index[1].spectrum.put(np.array([exposure_time * 50e3] * npts))

    # fake data for parsing
    gather_client = fly1d.gather_client
    gather_client.gathered = make_1d_data(startx, endx, npts,
                                          gathered_points=100,
                                          gate_enable=fly1d._gate_enable)


def test_fly1d(fly1d, sim_det, ipython, positioners):
    setup_scan(fly1d, ipython, positioners, configure=True)
    data = fly1d.run_subscan(0)
    df = pd.DataFrame(list(pd.DataFrame(data)['data']))
    print('acquired data:')
    print(df)

    for sig, value in sim_det.stage_sigs.items():
        print('det', sig.name, value, sig.setpoint_pvname)
    for sig, value in sim_det.cam.stage_sigs.items():
        print('cam', sig.name, value)
    for sig, value in sim_det.tiff1.stage_sigs.items():
        print('tiff1', sig.name, value)

    has_test_det = any(isinstance(det, TestDetector)
                       for det in fly1d.detectors)

    if has_test_det:
        assert 'sim_tiff' in df

        for img_uid in df['sim_tiff']:
            print(img_uid, filestore.api.retrieve(img_uid).shape)

    print(fly1d.collect())
    print(fly1d.describe())


def test_flyplan1d(positioners, fly1d, run_engine):
    import hxnfly.bs
    plan = hxnfly.bs.FlyPlan1D()
    gen = plan(positioners['testx'], -1.0, 1.0, 2, 1.0, dead_time=0.001)
    run_engine(gen)


def test_flyplan1d_with_callbacks(request, positioners, fly1d, run_engine):
    import hxnfly.bs
    from hxnfly.callbacks import FlyLivePlot
    #    , FlyRoiPlot, FlyLiveImage)
    from hxnfly.callbacks import FlyDataCallbacks
    plan = hxnfly.bs.FlyPlan1D()

    # flyplot = FlyRoiPlot(['Pt'], channels=[1, 2, 3], use_sum=True)
    # fly2dplot = FlyLiveImage(['As','Ba','S','Cs','Ti'], channels=[1, 2, 3],
    #                          use_sum=True)
    # fly2dplot1 = FlyLiveCrossSection(['Ba'], channels=[1, 2, 3],
    #                                  use_sum=True)
    point_signal = MockSignal('point_signal', 3)
    data_signal1 = MockSignal('data_signal1', [1, 0.5, 2])
    data_signal2 = MockSignal('data_signal2', [2, 1.5, 1])
    liveplot = FlyLivePlot({'sig1': [data_signal1],
                            'sig2': [data_signal2]},
                           point_signal=point_signal)
    plan.subs = [FlyDataCallbacks(), liveplot]
    gen = plan(positioners['testx'], -1.0, 1.0, 2, 1.0, dead_time=0.001)
    run_engine(gen)

    plt.savefig('liveplot-{}.png'.format(request.node.name))
    liveplot.disable()
    plt.clf()


def test_fly1d_plot_position(request, fly1d, sim_det, ipython, positioners):
    setup_scan(fly1d, ipython, positioners, configure=True)
    fly1d.run_subscan(0)

    from hxnfly.plot import plot_position
    plot_position(fly1d.scan_id, fly1d.data[fly1d.scan_id], fly1d._debug_data,
                  left='testx', right='testy')
    plt.savefig('plot_position-{}.png'.format(request.node.name))
    plt.clf()


def test_mll_imports(ipython):
    import hxnfly.fly_mll


def test_zp_imports(ipython):
    import hxnfly.fly_zp


def test_mllzp_imports(ipython):
    import hxnfly.hxn_fly



import os
import logging
import logging.handlers
from . import logger


handler = None


def log_setup(level=logging.DEBUG, log_path=None):
    global handler

    if handler is not None:
        return

    if log_path is None:
        log_path = os.path.expanduser('~/flyscan_debug.log')

    fmt = logging.Formatter('%(asctime)s - %(funcName)s/%(levelname)s: '
                            '%(message)s')

    handler = logging.handlers.TimedRotatingFileHandler(log_path, 'W6')
    handler.setLevel(level)

    handler.setFormatter(fmt)
    logger.addHandler(handler)


def log_time(logger, time_dict, name, *, t1=None, t2=None, ms=True):
    '''Calculate time difference and log it
    '''
    if t1 is not None and t2 is not None:
        dt = time_dict[t2] - time_dict[t1]
        time_dict[name] = dt
    else:
        dt = time_dict[name]

    if dt > 5.0 and ms:
        ms = False

    if ms:
        units = 'ms'
        dt *= 1000.0
    else:
        units = 's'

    logger.debug('%s: %.1f %s', name, dt, units)



from __future__ import print_function
import logging

import os
import numpy as np
import IPython
from ppmac.pp_comm import PPComm


logger = logging.getLogger(__name__)


def ppmac_connect(force=False):
    """
    Store the ppmac connection in the user namespace.
    Set environment variable ``PPMAC_HOST`` to valid host name or IP address.
    """
    ip = IPython.get_ipython()
    ppmac = None
    if force or '_ppmac' not in ip.user_ns:
        logger.debug('Connecting to the Power PMAC...')
        try:
            host = os.environ.get('PPMAC_HOST', None)
            if not host:
                raise ValueError("Host name is not set: environment variable 'PPMAC_HOST' is not set")

            ppmac = ip.user_ns['_ppmac'] = PPComm(host=host,
                                                  fast_gather=True)
        except Exception as ex:
            logger.error('Failed to connect to Power PMAC: %s', ex,
                         exc_info=ex)
        except KeyboardInterrupt:
            logger.error('Failed to connect to Power PMAC'
                         '(interrupted by user)')
        else:
            logger.debug('Connected')
    else:
        logger.debug('Already connected to the ppmac')
        ppmac = ip.user_ns['_ppmac']

    ip.user_ns['_ppmac'] = ppmac
    return ppmac


def filter_ppmac_data(servo_period, servo_t, enable, motor_cols, other_cols,
                      axis_names, axis_keys, have_target_pos, gate_enable,
                      home_pos):
    '''Filter gathered power pmac data, returning two sets of data:
    scan_data, debug_data

    Where scan_data contains normal columns you expect to see in a table (such
    as timestamp, [average] motor position in one frame, detector alive time,
    detector dead time - according to ppmac servo clock)

    And debug_data has high-frequency servo data, target position information,
    and other stuff that may be useful for debugging/plotting (see plot.py)
    '''
    # -- time --
    servo_t *= servo_period
    servo_t -= servo_t[0]

    # -- enable signal --
    enable = enable.astype(int)

    diff = np.diff(enable)

    if gate_enable == 'low_to_high':
        start_i, = np.where(diff == 1)
        end_i, = np.where(diff == -1)
    elif gate_enable == 'high_to_low':
        start_i, = np.where(diff == -1)
        end_i, = np.where(diff == 1)

    slices = [slice(start, end) for start, end in zip(start_i, end_i)]
    mid_t = [np.average(servo_t[slice_]) for slice_ in slices]

    start_t = servo_t[start_i]
    end_t = servo_t[end_i]

    alive_t = [en - st for st, en in zip(start_t, end_t)]
    dead_t = [st - en for st, en in zip(start_t[1:], end_t[:-1])]
    dead_t.append(servo_t[-1] - end_t[-1])

    # -- motor positions --
    axis_names = list(sorted(axis_names))
    gathered_pos = [(motor_cols[i] - home_pos[axis_name])
                    for i, axis_name in enumerate(axis_names)
                    ]

    if have_target_pos:
        target_pos = {axis_key: (other_cols[i] - home_pos[axis_name])
                      for i, (axis_key, axis_name) in
                      enumerate(zip(axis_keys, axis_names))
                      }

    average_pos = [[np.average(pos[slice_]) for slice_ in slices]
                   for pos in gathered_pos]

    # -- return it all in a dictionary --
    logger.debug('total scaler alive time: %.1f', np.sum(alive_t))
    logger.debug('total scaler dead time: %.1fs', np.sum(dead_t))
    logger.debug('scaler alive+dead sum: %.1fs', np.sum(alive_t) +
                 np.sum(dead_t))

    data = dict(elapsed_time=mid_t,
                alive=alive_t,
                dead=dead_t,
                )

    debug_data = dict(start_time=start_t,
                      end_time=end_t,
                      servo_time=servo_t,
                      servo_enable=enable,
                      )

    items = (axis_keys, gathered_pos, average_pos)
    for key, gathered, average in zip(*items):
        data[key] = average

        debug_data['servo_{}'.format(key)] = gathered
        if have_target_pos:
            debug_data['target_{}'.format(key)] = target_pos[key]

    return data, debug_data


def gather_setup(axis_numbers, *, period=2, additional_addresses=None,
                 include_target_positions=False, gpascii=None,
                 send_to_controller=False):
    if additional_addresses is None:
        additional_addresses = []

    start_lines = ['gather.enable=0']

    addresses = ['Sys.ServoCount.a',
                 'Sys.M[100].a',
                 ]

    for axis_num in axis_numbers:
        addresses.append('Motor[{}].Pos.a'.format(axis_num))

    if include_target_positions:
        for axis_num in axis_numbers:
            addresses.append('Motor[{}].Desired.Pos.a'.format(axis_num))

    addresses.extend(additional_addresses)

    addr_lines = ['gather.addr[{}]={}'.format(i, addr)
                  for i, addr in enumerate(addresses)]

    end_lines = ['gather.items={}'.format(len(addresses)),
                 'gather.period={}'.format(period),
                 'gather.enable=1',
                 'gather.enable=0',
                 'gather.MaxSamples=Gather.MaxLines',
                 ]

    logger.debug('Sending gather configuration ({} addresses):'
                 ''.format(len(addresses)))

    config = start_lines + addr_lines + end_lines
    if send_to_controller:
        if gpascii is None:
            raise ValueError('gpascii must be specified to send to controller')

        for line in config:
            logger.debug('> ' + line)
            gpascii.send_line(line)

        gpascii.sync()

    return config


def is_closed_loop(gpascii, axis):
    var = 'Motor[{}].ClosedLoop'
    return gpascii.get_variable(var.format(axis), type_=int)



from __future__ import print_function
import numpy as np


def get_mds_info(positioners=None, detectors=None, data=None,
                 desc=None):
    """Helper function to extract information from the positioners/detectors
    and send it over to metadatastore so that ophyd is insulated from mds
    spec changes

    Parameters
    ----------
    positioners : list, optional
        List of ophyd positioners
    detectors : list
        List of ophyd detectors, optional
    data : dict
        Dictionary of actual data
    """
    if desc is None:
        desc = {}

    [desc.update(x.describe()) for x in (detectors + positioners)]

    info_dict = {}
    for name, value in data.iteritems():
        # grab 'value' from [value, timestamp]
        val = np.asarray(value[0])

        dtype = 'number'
        try:
            shape = val.shape
        except AttributeError:
            # val is probably a float...
            shape = None

        if shape:
            dtype = 'array'

        d = {'dtype': dtype, 'shape': shape}
        d.update(desc[name])
        d = {name: d}
        info_dict.update(d)

    return info_dict
